#! /usr/bin/perl
################################################################################
#Copyright (C) 2000  by Christopher Lansdown
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##############################################################################
#########
#Modules
use Gtk;
use Checkbook;

#########################
#Configuration Variables
my $icon_dir = '/usr/local/share/checkbook';

################
#Initialize Gtk
init Gtk;

parse Gtk::Rc($ENV{HOME} . '/.checkbookrc');

#################
#Local Variables
my $checkbook = Checkbook::new();
my $window;
my $vbox;
my $scrolled_win;
my $clist;
my $toolbar;
my $arguments = {};
my $clists = [];
my $button;
my $i;
my $accel_group;
my $item_factory;

########################
#Create our main window
$window = new Gtk::Window('toplevel');
$window->set_title('Checkbook');
$window->signal_connect( 'destroy', sub { Gtk::main_quit( Gtk ); } );
$window->set_policy(1, 1, 1);
realize $window;

####################
#Make our main vbox
$vbox = new Gtk::VBox(0, 0);
$window->add($vbox);
show $vbox;

######################
#Set up the arguments
$arguments->{checkbook} = $checkbook;
$arguments->{clists} = $clists;
$arguments->{notebook} = create_notebook($arguments);


#################
#Create the menu
$accel_group = new Gtk::AccelGroup;
$item_factory = new Gtk::ItemFactory('Gtk::MenuBar', "<main>", $accel_group);
$accel_group->attach($window);
$item_factory->create_item(['/File', undef, 0, '<Branch>'], \&menu_callback);
$item_factory->create_item(['/File/_Save', '<control>S', 1], \&menu_callback, $arguments);
$item_factory->create_item(["/File/sep1", undef, 0, "<Separator>"], \&menu_callback, $arguments);
$item_factory->create_item(['/File/_Quit', '<control>Q', 2], sub { Gtk::main_quit(Gtk); });

$item_factory->create_item(['/Operations', undef, 0, '<Branch>'], \&menu_callback, $arguments);
$item_factory->create_item(['/Operations/_Insert', '<control>I', 20], \&menu_callback, $arguments);
$item_factory->create_item(['/Operations/_Delete', '<control>D', 21], \&menu_callback, $arguments);
$item_factory->create_item(['/Operations/_Edit', '<control>E', 22], \&menu_callback, $arguments);
$item_factory->create_item(['/Operations/Create Register', undef, 23], \&menu_callback, $arguments);
$item_factory->create_item(['/Operations/Delete Register', undef, 24], \&menu_callback, $arguments);

$item_factory->create_item(['/Preferences', undef, 0, '<Branch>'], \&menu_callback, $arguments);
$item_factory->create_item(['/Preferences/_Configuration', '<control>C', 30], \&menu_callback, 
			   $arguments);

$item_factory->create_item(["/Help",	undef,	0,	"<LastBranch>"], \&menu_callback, $arguments);
$item_factory->create_item(["/Help/Getting Started", undef, 40], \&menu_callback, $arguments);
$item_factory->create_item(["/Help/License", undef, 41], \&menu_callback, $arguments);
$item_factory->create_item(["/Help/About", undef, 49], \&menu_callback, $arguments);


$vbox->pack_start($item_factory->get_widget('<main>'), 0, 0, 0);

####################
#Set up our toolbar
$toolbar = new Gtk::Toolbar('horizontal', 'both');
$toolbar->set_space_style('line');
$toolbar->set_button_relief("none");
$button = $toolbar->append_item ( "Insert", "Insert an entry into the checkbook",
		"Toolbar/Insert", 
new_pixmap("$icon_dir/insert.xpm",$window->window, $window->style->bg('normal')));
	$button->signal_connect('clicked', \&insert_window, $arguments);

$button = $toolbar->append_item ( "Delete", "Delete an entry from the checkbook",
		"Toolbar/Delete", 
new_pixmap("$icon_dir/delete.xpm",$window->window, $window->style->bg('normal')));
	$button->signal_connect('clicked', \&delete_entry, $arguments);

$button = $toolbar->append_item ( "Edit", "Edit an entry from in checkbook",
		"Toolbar/Edit", 
new_pixmap("$icon_dir/edit.xpm",$window->window, $window->style->bg('normal')));
	$button->signal_connect('clicked', \&edit_window, $arguments);

$toolbar->append_space();

$button = $toolbar->append_item ( "Save", "Save the checkbook",
		"Toolbar/Save",
new_pixmap("$icon_dir/save.xpm",$window->window, $window->style->bg('normal')));
	$button->signal_connect('clicked', \&save_checkbook, $arguments);

$toolbar->append_space();

$button = $toolbar->append_item ( "Quit", "Quit",
		"Toolbar/Quit",
new_pixmap("$icon_dir/quit.xpm", $window->window, $window->style->bg('normal')));
	$button->signal_connect('clicked', sub { Gtk::main_quit( Gtk ); });

show $toolbar;
$vbox->pack_start($toolbar, 0, 1, 0);

#####################
#Add in the notebook
$arguments->{notebook}->show;
$vbox->pack_start($arguments->{notebook}, 1, 1, 0);

##################
#Display the list
for($i = 0; $i < scalar(@{$checkbook->{names}}); $i++) {
      display_checkbook($checkbook->{checkbooks}[$i], 
			$arguments->{clists}[$i]{clist},
			$checkbook->{prefs});
}

################################################
#Show the main window and go into the main loop
$window->set_usize($checkbook->{prefs}->{xsize},
		   $checkbook->{prefs}->{ysize});
$window->show_all();
Gtk::main(Gtk);


sub edit_window
{
      my $button = shift;
      my $argument = shift;
      my $checkbook = $argument->{checkbook};
      my $index = $argument->{notebook}->get_current_page;
      my $register = $checkbook->{checkbooks}[$index];
      my $clist_row;
      
      if($argument->{clist_row} < 0) {
	    return;
      } else {
	    $clist_row = $argument->{clists}[$index]{clist_row};
      }

      my $number = $register->[$clist_row]->{number};
      my $date = $register->[$clist_row]->{date};
      my $description = $register->[$clist_row]->{description};
      my $payment = $register->[$clist_row]->{payment};
      my $code = $register->[$clist_row]->{code};
      my $fee = $register->[$clist_row]->{fee};
      my $deposit = $register->[$clist_row]->{deposit};
      
      my $elements = entry_window($argument, 
				  $number, 
				  $date, 
				  $description, 
				  $payment, 
				  $code, 
				  $fee, 
				  $deposit);
      $elements->{button}->signal_connect('clicked', \&edit_entry, $elements);
      $elements->{window}->show_all;
      $elements->{clist} = $argument->{clists}[$index]->{clist};
      $elements->{clist_row} = $clist_row;
      $elements->{checkbook} = $checkbook;
      $elements->{register} = $register;
      $elements->{index} = $index;

      1;
}

sub edit_entry
{
      my $button = shift;
      my $element = shift;
      
      my %values;
      $values{number} = $element->{number}->get_text();
      $values{date} = sprintf("%.2d", $element->{year}->get_text()) . '/' .
	sprintf("%.2d", $element->{month}->get_text()) . '/' .
	  sprintf("%.2d", $element->{day}->get_text());
      $values{description} = $element->{description}->get_text();
      $values{payment} = $element->{payment}->get_text();
      $values{code} = $element->{code}->get_text();
      $values{fee} = $element->{fee}->get_text();
      $values{deposit} = $element->{deposit}->get_text();

      $element->{checkbook}->replace($element->{index}, $element->{clist_row}, \%values);

      display_checkbook($element->{register}, $element->{clist}, 
			$element->{checkbook}->{prefs});

      $element->{window}->destroy;

      1;
}

sub save_checkbook
{
      my $button = shift;
      my $argument = shift;

      $argument->{checkbook}->save();


}

sub select_clist {
	my($widget, $data, $row, $column, $event) = @_;
	$data->{arguments}->{clists}[$data->{index}]{clist_row} = $row;

	$row;
}

sub unselect_clist {
	my( $widget, $data, $row, $column, $event) = @_;
	if($data->{arguments}->{clists}[$data->{index}]{clist_row} == $row) {
	      $data->{arguments}->{clists}[$data->{index}]{clist_row} = -1;
	}

	$row;
}

#
#The data to this function should be an array of the form
#[length, $next_widget]
sub move_next {
      my ($entry, $data) = @_;
      my $length = length($entry->get_text());
      if($length >= $$data[0]) {
	    $$data[1]->grab_focus();
#	    $$data[1]->select_region(0, $length);
      }
      0;
}

sub entry_window
{
      my $argument = shift;
      my $number = shift;
      my $date = shift;
      my $description = shift;
      my $payment = shift;
      my $code = shift;
      my $fee = shift;
      my $deposit = shift;

      my $element = {};
      my $window;
      my $hbox;
      my $vbox;
      my $label;
      my $entry;
      my $year;
      my $month;
      my $day;
      

      $element->{clist} = $$argument{clist};
      $element->{checkbook} = $$argument{checkbook};

      $window = new Gtk::Window('toplevel');
      $window->set_title('Insert an Entry');
      $window->position(-mouse);
      $element->{window} = $window;

      $vbox = new Gtk::VBox(0, 0);
      $window->add($vbox);

      ################
      #The first hbox
      $hbox = new Gtk::HBox(0, 5);
      $vbox->pack_start($hbox, 0, 0, 0);

      ##################
      #The check number
      $label = new Gtk::Label('Check #: ');
      $hbox->pack_start($label, 0, 0, 0);

      $entry = new Gtk::Entry();
      $entry->set_max_length(4);
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "99999") + 4, -1);
      if($number) { $entry->set_text($number); }
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{number} = $entry;

      ##########
      #The date
      $date =~ m|(....)/(..)/(..)|;
      $year = $1;
      $month = $2;
      $day = $3;

      #year
      $label = new Gtk::Label('Year: ');
      $hbox->pack_start($label, 0, 0, 0);

      $entry = new Gtk::Entry();
      $entry->set_max_length(4);
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "20000"), -1);
      if($year) { $entry->set_text($year); }
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{year} = $entry;
      $entry->signal_connect('focus_in_event', sub {
				   $_[0]->set_position(0); 
				   $_[0]->select_region(0, 4); 
				   return 1;}, $entry);
      $entry->signal_connect('focus_out_event', sub {$_[0]->select_region(0, 0); }, $entry);
      $entry->signal_connect('button_press_event', sub {$_[0]->signal_emit_stop_by_name('button_press_event'); $_[0]->grab_focus(); 1; }, undef);
      $element->{number}->signal_connect('activate', 
				       \&move_next, [0, $entry]);

      #month
      $label = new Gtk::Label('Month: ');
      $hbox->pack_start($label, 0, 0, 0);

      $entry = new Gtk::Entry();
      $entry->set_max_length(2);
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "120"), -1);
      if($month) { $entry->set_text($month); }
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{month} = $entry;
      $entry->signal_connect('focus_in_event', sub {$_[0]->set_position(0); $_[0]->select_region(0, 2); }, $entry);
      $entry->signal_connect('focus_out_event', sub {$_[0]->select_region(0, 0); }, $entry);
      $entry->signal_connect('button_press_event', sub {$_[0]->signal_emit_stop_by_name('button_press_event'); $_[0]->grab_focus(); 1; }, undef);
      $element->{year}->signal_connect('changed', 
				       \&move_next, [4, $entry]);
      $element->{year}->signal_connect('activate', 
				       \&move_next, [0, $entry]);

      
      #day
      $label = new Gtk::Label('Day: ');
      $hbox->pack_start($label, 0, 0, 0);

      $entry = new Gtk::Entry();
      $entry->set_max_length(2);
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "300") + 4, -1);
      if($day) { $entry->set_text($day); }
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{day} = $entry;
      $entry->signal_connect('focus_in_event', sub {$_[0]->set_position(0); $_[0]->select_region(0, 2); }, $entry);
      $entry->signal_connect('focus_out_event', sub {$_[0]->select_region(0, 0); }, $entry);
      $entry->signal_connect('button_press_event', sub {$_[0]->signal_emit_stop_by_name('button_press_event'); $_[0]->grab_focus(); 1; }, undef);
      $element->{month}->signal_connect('changed', 
				       \&move_next, [2, $entry]);
      $element->{month}->signal_connect('activate', 
				       \&move_next, [0, $entry]);

      
      ##################
      #The description
      $label = new Gtk::Label('Description: ');
      $label->show;
      $hbox->pack_start($label, 0, 0, 0);
      
      $entry = new Gtk::Entry();
      if($description) { $entry->set_text($description); }
      $entry->show;
      $hbox->pack_start($entry, 1, 1, 0);
      $element->{description} = $entry;
      $element->{day}->signal_connect('changed', 
				       \&move_next, [2, $entry]);
      $element->{day}->signal_connect('activate', 
				       \&move_next, [0, $entry]);



      #################
      #The second hbox
      $hbox = new Gtk::HBox(0, 5);
      $hbox->show;
      $vbox->pack_start($hbox, 0, 0, 0);

     ##############
      #The Payment
      $label = new Gtk::Label('Payment/Debit: ');
      $label->show;
      $hbox->pack_start($label, 0, 0, 0);
      
      $entry = new Gtk::Entry();
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "1,000,000.00 ") + 4, -1);
      if($payment) { $entry->set_text($payment); }
      $entry->show;
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{payment} = $entry;
      $element->{description}->signal_connect('activate', 
				       \&move_next, [0, $entry]);


      ##########
      #The Code
      $label = new Gtk::Label('Code: ');
      $label->show;
      $hbox->pack_start($label, 0, 0, 0);

      $entry = new Gtk::Entry();
      $entry->set_max_length(4);
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "99999") + 4, -1);
      if($code) { $entry->set_text($code); }
      $entry->show;
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{code} = $entry;
      $element->{payment}->signal_connect('activate', 
					  \&move_next, [0, $entry]);

      #########
      #The Fee
      $label = new Gtk::Label('Fee: ');
      $label->show;
      $hbox->pack_start($label, 0, 0, 0);

      $entry = new Gtk::Entry();
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "99,999.99") + 4, -1);
      if($fee) { $entry->set_text($fee); }
      $entry->show;
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{fee} = $entry;
      $element->{code}->signal_connect('activate', 
				       \&move_next, [0, $entry]);
      
      ##############
      #The Deposit
      $label = new Gtk::Label('Deposit/Credit: ');
      $label->show;
      $hbox->pack_start($label, 0, 0, 0);
      
      $entry = new Gtk::Entry();
      $entry->set_usize(Gtk::Gdk::Font::string_width($window->style->font, "1,000,000.00 ") + 4, -1);
      if($deposit) { $entry->set_text($deposit); }
      $entry->show;
      $hbox->pack_start($entry, 0, 0, 0);
      $element->{deposit} = $entry;
      $element->{fee}->signal_connect('activate', 
				       \&move_next, [0, $entry]);

      

      my $button = new Gtk::Button(' Insert ');
      $element->{button} = $button;
      $button->show;
      $hbox->pack_start($button, 0, 0, 0);
      $element->{deposit}->signal_connect('activate', 
					  \&move_next, [0, $button]);

      $button = new Gtk::Button(' Cancel ');
      $button->show;
      $hbox->pack_start($button, 0, 0, 0);
      $button->signal_connect('clicked', sub { $window->destroy; }, undef);

      return $element;
}

sub insert_window
{
      my $button = shift;
      my $arguments = shift;
      my $element;

      my @time = localtime;
      my $time = sprintf("%.4d/%.2d/%.2d", $time[5] + 1900, $time[4] + 1, $time[3]);

      $element = entry_window($arguments, undef, $time);
      $element->{clist} = $arguments->{clists}[$arguments->{notebook}->get_current_page]->{clist};
      $element->{index} = $arguments->{notebook}->get_current_page;
      $element->{checkbook} = $arguments->{checkbook};
      $element->{button}->signal_connect('clicked', \&insert_entry, $element);
      $element->{window}->show_all;
}

sub insert_entry
{
      my $button = shift;
      my $element = shift;
      my $index = $element->{index};
      my $checkbook = $element->{checkbook};
      my $clist = $element->{clist};
      my $number = $element->{number}->get_text();
      my $date = sprintf("%.4d", $element->{year}->get_text()) . '/' .
	sprintf("%.2d", $element->{month}->get_text()) . '/' .
	  sprintf("%.2d", $element->{day}->get_text());
#      my $date = $element->{date}->get_text();
      my $description = $element->{description}->get_text();
      my $payment = $element->{payment}->get_text();
      my $code = $element->{code}->get_text();
      my $fee = $element->{fee}->get_text();
      my $deposit = $element->{deposit}->get_text();

      ##################
      #Insert the entry
      $checkbook->insert($index, $number, $date, $description, $payment, $code, $fee, $deposit);

      $element->{window}->destroy();

      ##############################
      #Then redisplay the checkbook
      display_checkbook($element->{checkbook}->{checkbooks}[$index], $clist,
		       $checkbook->{prefs});

      return 1;
}

sub delete_entry
{
      my $button = shift;
      my $arguments = shift;
      
      if($arguments->{clist_row} < 0) {
	    return 1;
      } else {
	    my $index = $arguments->{notebook}->get_current_page;
	    $arguments->{checkbook}->remove($index, $arguments->{clists}[$index]{clist_row});
	    display_checkbook($arguments->{checkbook}->{checkbooks}[$index], 
			      $arguments->{clists}[$index]{clist},
			      $arguments->{checkbook}->{prefs});
      }


      1;
}

#syntax:
#display_checkbook($checkbook_list_ref, $clist);
sub display_checkbook
{
      my $checkbook = shift;
      my $clist = shift;
      my $prefs = shift;
      my $entry;

#      print "There are " . scalar(@$checkbook) . " entries in the checkbook.\n";

      my ($number, $date, $description, $payment, $code, $fee, $deposit);

      my ($year, $month, $day);

      $clist->freeze();
      $clist->clear();
      foreach $entry (@$checkbook) {
	    ################################
	    #Skip over non-existent entries
	    if(!defined($entry)) { next; }
	    if($entry->{number} == 0 and $entry->{date} eq '' and $entry->{description} eq '' and $entry->{payment} == 0 and $entry->{deposit} == 0 and $entry->{fee} == 0 and $entry->{code} == 0) {
		  next;
	    }

	    #################
	    #Format properly
	    if($entry->{number} ne '') { $number = sprintf("%.4d", $entry->{number}); } else { $number = ''; }
	    if($entry->{payment} ne '') { $payment = sprintf("%.2f", $entry->{payment}); } else { $payment = ''; }
	    if($entry->{deposit} ne '') { $deposit = sprintf("%.2f", $entry->{deposit}); } else { $deposit = ''; }
	    if($entry->{fee} ne '') { $fee = sprintf("%.2f", $entry->{fee}); } else { $fee = ''; }
	    $entry->{date} =~ /(....).(..).(..)/;
	    $year = $1;
	    $month = $2;
	    $day = $3;
	    if($prefs->{date} eq 'MDY') {
		  $date = "$month/$day/$year";
	    } elsif($prefs->{date} eq 'MYD') {
		  $date = "$month/$year/$day";
	    } elsif($prefs->{date} eq 'DMY') {
		  $date = "$day/$month/$year";
	    } elsif($prefs->{date} eq 'DYM') {
		  $date = "$day/$year/$month";
	    } elsif($prefs->{date} eq 'YDM') {
		  $date = "$year/$day/$month";
	    } elsif($prefs->{date} eq 'DM') {
		  $date = "$day/$month";
	    } elsif($prefs->{date} eq 'MD') {
		  $date = "$month/$day";
	    } else {
		  $date = $entry->{date};
	    }


	    #######################
	    #Do the actual display
	    my @strings = ($number,
			   $date, 
			   $entry->{description},
			   $payment,
			   $entry->{code},
			   $fee,
			   $deposit,
			   sprintf("%.2f", $entry->{balance})
			  );
	    $clist->append(@strings);
      }
      $clist->thaw();

      return;
}

sub new_pixmap {
	my ($filename, $window, $background) = @_;
	my ($pixmap, $mask) = create_from_xpm Gtk::Gdk::Pixmap($window, $background, $filename);


	return new Gtk::Pixmap($pixmap, $mask);
}

#syntax:
#$notebook_widget = create_notebook($argument_hash_ref)
sub create_notebook {
      my $arguments = shift;
      my $checkbook = $arguments->{checkbook};

      my @titles = ('Check #', 'Date', 'Description', 'Payment/Debit', 'Code', 'Fee', 
		    'Deposit/Credit', 'Balance');
      my $notebook;
      my $scrolled_window;
      my $clist;
      my $label;
      my $i;

      $notebook = new Gtk::Notebook;
      $notebook->popup_enable;

      foreach my $name (@{$checkbook->{names}}) {
	    ##################
	    #Create the clist
	    $scrolled_win = new Gtk::ScrolledWindow(undef, undef);
	    $scrolled_win->set_policy('automatic', 'automatic');
	    $clist = new_with_titles Gtk::CList(@titles);
	    
	    ####################
	    #Configure the list
	    $clist->set_column_justification(0, 'right');		   
	    $clist->set_column_justification(3, 'right');		   
	    $clist->set_column_justification(4, 'right');		   
	    $clist->set_column_justification(5, 'right');		   
	    $clist->set_column_justification(6, 'right');		   
	    $clist->set_column_justification(7, 'right');		   
	    
	    ###################
	    #Set up the widths
	    #the description for column 1 is wide enough
	    $clist->set_column_width(1, Gtk::Gdk::Font::string_width($clist->style->font, 
								     "2000/12/300"));
	    $clist->set_column_width(2, Gtk::Gdk::Font::string_width($clist->style->font, 
								     "This is a generic description."));
	    #the description for column 3 is wide enough
	    $clist->set_column_width(4, Gtk::Gdk::Font::string_width($clist->style->font, "0000"));
	    $clist->set_column_width(5, Gtk::Gdk::Font::string_width($clist->style->font, "129.95"));
	    #the description for column 6 is wide enough
	    $clist->set_column_width(7, Gtk::Gdk::Font::string_width($clist->style->font, 
								     "1,000,000.00"));
	    
	    ###############################
	    #Setup other stuff on the list
	    $clist->signal_connect('select_row', \&select_clist, {'arguments', $arguments,
								  'index', $i});
	    $clist->signal_connect('unselect_row', \&unselect_clist, {'arguments', $arguments,
								      'index', $i});
	    $clist->border_width(5);
	    $scrolled_win->add($clist);
	    $clist->show;
	    $scrolled_win->show;

	    $label = new Gtk::Label($name);
	    $label->show;
	    
	    $notebook->append_page($scrolled_win, $label);
	    $arguments->{clists}[$i] = {};
	    $arguments->{clists}[$i]->{clist} = $clist;
	    $arguments->{clists}[$i]->{clist_row} = -1;
	    $i++;
      }

      return $notebook;
}

sub create_register_window {
      my $widget = shift;
      my $arguments = shift;

      my $window = new Gtk::Window('toplevel');
      my $hbox = new Gtk::HBox(0, 0);
      $window->add($hbox);
      $window->position(-mouse);

      my $label = new Gtk::Label('Name: ');
      $hbox->pack_start($label, 0, 0, 0);
      my $input = new Gtk::Entry();
      $hbox->pack_start($input, 0, 0, 0);
      my $button = new Gtk::Button('Create');
      $button->signal_connect('clicked', \&create_register, {'window', $window, 
							     'input', $input,
							    'arguments', $arguments});
      $input->signal_connect('activate', \&create_register, {'window', $window, 
							     'input', $input,
							    'arguments', $arguments});
      $hbox->pack_start($button, 0, 0, 0);
      $button = new Gtk::Button('Cancel');
      $button->signal_connect('clicked', sub { $window->destroy }, $window);
      $hbox->pack_start($button, 0, 0, 0);
      

      $window->show_all;
      $input->grab_focus();
}

sub create_register {
      my $button = shift;
      my $elements = shift;
      my $arguments = $elements->{arguments};

      my $name = $elements->{input}->get_text();
      my $scrolled_win;
      my $clist;
      my $label;
      my @titles = ('Check #', 'Date', 'Description', 'Payment/Debit', 'Code', 'Fee', 
		    'Deposit/Credit', 'Balance');
    
      $arguments->{checkbook}->new_register($name);

      $elements->{window}->destroy;


      ##################
      #Create the clist
      $scrolled_win = new Gtk::ScrolledWindow(undef, undef);
      $scrolled_win->set_policy('automatic', 'automatic');
      $clist = new_with_titles Gtk::CList(@titles);
      
      ####################
      #Configure the list
      $clist->set_column_justification(0, 'right');		   
      $clist->set_column_justification(3, 'right');		   
      $clist->set_column_justification(4, 'right');		   
      $clist->set_column_justification(5, 'right');		   
      $clist->set_column_justification(6, 'right');		   
      $clist->set_column_justification(7, 'right');		   
      
      ###################
      #Set up the widths
      #the description for column 1 is wide enough
      $clist->set_column_width(1, Gtk::Gdk::Font::string_width($clist->style->font, 
							       "2000/12/300"));
      $clist->set_column_width(2, Gtk::Gdk::Font::string_width($clist->style->font, 
							       "This is a generic description."));
      #the description for column 3 is wide enough
      $clist->set_column_width(4, Gtk::Gdk::Font::string_width($clist->style->font, "0000"));
      $clist->set_column_width(5, Gtk::Gdk::Font::string_width($clist->style->font, "129.95"));
      #the description for column 6 is wide enough
      $clist->set_column_width(7, Gtk::Gdk::Font::string_width($clist->style->font, 
							       "1,000,000.00"));
      
      ###############################
      #Setup other stuff on the list
      $clist->signal_connect('select_row', \&select_clist, {'arguments', $arguments,
							    'index', $i});
      $clist->signal_connect('unselect_row', \&unselect_clist, {'arguments', $arguments,
								'index', $i});
      $clist->border_width(5);
      $scrolled_win->add($clist);
      $clist->show;
      $scrolled_win->show;
      
      $label = new Gtk::Label($name);
      $label->show;
      
      $arguments->{notebook}->append_page($scrolled_win, $label);

      my $i = scalar(@{$arguments->{clists}});

      $arguments->{clists}[$i] = {};
      $arguments->{clists}[$i]->{clist} = $clist;
      $arguments->{clists}[$i]->{clist_row} = -1;
      
      1;
}


sub delete_register_window {
      my $widget = shift;
      my $arguments = shift;

      my $window = new Gtk::Window('toplevel');
      my $hbox = new Gtk::HBox(0, 0);
      $window->add($hbox);
      $window->position(-mouse);

      my $label = new Gtk::Label('Register #: ');
      $hbox->pack_start($label, 0, 0, 0);
      my $input = new Gtk::Entry();
      $hbox->pack_start($input, 0, 0, 0);
      my $button = new Gtk::Button('Delete');
      $button->signal_connect('clicked', \&delete_register, {'window', $window, 
							     'input', $input,
							     'arguments', $arguments});
      $hbox->pack_start($button, 0, 0, 0);
      $button = new Gtk::Button('Cancel');
      $button->signal_connect('clicked', sub { $window->destroy }, $window);
      $hbox->pack_start($button, 0, 0, 0);
      

      $window->show_all;

      1;
}

sub delete_register {
      my $button = shift;
      my $elements = shift;
      my $arguments = $elements->{arguments};

      my $index = $elements->{input}->get_text();
      $arguments->{checkbook}->delete_register($index);
      $arguments->{notebook}->remove_page($index);

      $elements->{window}->destroy;

      1;
}


sub menu_callback {
      my($menu_item, $action, $data) = @_;

      if($action == 2) {
	    save_checkbook($menu_item, $data);
      } elsif($action == 20) {
	    insert_window($menu_item, $data);
      } elsif ($action == 21) {
	    delete_entry($menu_item, $data);
      } elsif ($action == 22) {
	    edit_window($menu_item, $data);
      } elsif ($action == 23) {
	    create_register_window($menu_item, $data);
      } elsif ($action == 24) {
	    delete_register_window($menu_item, $data);
      } elsif ($action == 30) {
	    preferences_config($menu_item, $data);
      } elsif ($action == 40) {
	    getting_started_box($menu_item, $data);
      } elsif ($action == 41) {
	    license_box($menu_item, $data);
      } elsif ($action == 49) {
	    about_box($menu_item, $data);
      }

      1;
}

sub preferences_config {
      my $button = shift;
      my $data = shift;

      my $prefs = $data->{checkbook}->{prefs};
      my $elements = {};

      $elements->{data} = $data;

      my ($window, $mainbox, $vbox, $hbox, $button, 
	  $frame, $radio1, $radio, $label);
      

      $window = new Gtk::Window('toplevel');
      $elements->{window} = $window;
      $window->set_position(-mouse);
      $window->realize;

      $mainbox = new Gtk::VBox(0, 0);
      $window->add($mainbox);

      $frame = new Gtk::Frame("Date Format");
      $mainbox->add($frame);
      $vbox = new Gtk::VBox(0, 0);
      $frame->add($vbox);

      $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 0, 0, 0);
      
      $radio = new Gtk::RadioButton('Year-Month-Day');
      $hbox->pack_start($radio, 0, 0, 0);
      $elements->{YMD} = $radio;
      $radio = new Gtk::RadioButton('Year-Day-Month', $radio);
      $elements->{YDM} = $radio;
      if($prefs->{date} eq 'YDM') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);
      $radio = new Gtk::RadioButton('Month-Day-Year', $radio);
      $elements->{MDY} = $radio;
      if($prefs->{date} eq 'MDY') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);
      $radio = new Gtk::RadioButton('Month-Day', $radio);
      $elements->{MD} = $radio;
      if($prefs->{date} eq 'MD') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);


      $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 0, 0, 0);
      $radio = new Gtk::RadioButton('Month-Year-Day', $radio);
      $elements->{MYD} = $radio;
      if($prefs->{date} eq 'MYD') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);
      $radio = new Gtk::RadioButton('Day-Month-Year', $radio);
      $elements->{DMY} = $radio;
      if($prefs->{date} eq 'DMY') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);
      $radio = new Gtk::RadioButton('Day-Year-Month', $radio);
      $elements->{DYM} = $radio;
      if($prefs->{date} eq 'DYM') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);
      $radio = new Gtk::RadioButton('Day-Month', $radio);
      $elements->{DM} = $radio;
      if($prefs->{date} eq 'DM') { $radio->set_active(1); }
      $hbox->pack_start($radio, 0, 0, 0);


      #######################
      #Set up the Size stuff
      $frame = new Gtk::Frame('Starting Size');
      $mainbox->pack_start($frame, 0, 0, 0);
      $vbox = new Gtk::VBox(0, 0);
      $frame->add($vbox);
      $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 0, 0, 0);
      $label = new Gtk::Label('X: ');
      $hbox->pack_start($label, 0, 0, 0);
      $entry = new Gtk::Entry();
      $elements->{xsize} = $entry;
      $hbox->pack_start($entry, 0, 0, 0);
      $entry->set_text($prefs->{xsize});
      $label = new Gtk::Label('Y: ');
      $hbox->pack_start($label, 0, 0, 0);
      $entry = new Gtk::Entry();
      $elements->{ysize} = $entry;
      $hbox->pack_start($entry, 0, 0, 0);
      $entry->set_text($prefs->{ysize});


      #####################
      #Miscellaneous stuff
      $frame = new Gtk::Frame('Miscellaneous');
      $mainbox->pack_start($frame, 0, 0, 0);
      $vbox = new Gtk::VBox(0, 0);
      $frame->add($vbox);
      $button = new Gtk::CheckButton('Sort entries at startup');
      if($prefs->{sortatstart} =~ /yes/io) { $button->set_active(1); }
      $vbox->pack_start($button, 0, 0, 0);
      $elements->{sortatstart} = $button;
      $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 0, 0, 0);
      $label = new Gtk::Label('Datafile: ');
      $hbox->pack_start($label, 0, 0, 0);
      $entry = new Gtk::Entry();
      $elements->{datafile} = $entry;
      $hbox->pack_start($entry, 1, 1, 1);
      $entry->set_text($prefs->{datafile});


      ###################
      #The functions bar
      $hbox = new Gtk::HBox(0, 0);
      $mainbox->pack_end($hbox, 0, 0, 0);
      $button = new Gtk::Button(' Help ');
      $button->signal_connect('clicked', \&config_help, $elements);
      $elements->{help} = $button;
      $hbox->pack_end($button, 0, 0, 1);
      $button = new Gtk::Button(' Close ');
      $button->signal_connect('clicked', sub { $window->destroy; }, $window);
      $hbox->pack_end($button, 0, 0, 1);
      $button = new Gtk::Button(' Apply ');
      $button->signal_connect('clicked', \&config_apply, $elements);
      $elements->{apply} = $button;
      $hbox->pack_end($button, 0, 0, 1);
      $button = new Gtk::Button(' Save ');
      $button->signal_connect('clicked', \&config_save, $elements);
      $elements->{save} = $button;
      $hbox->pack_end($button, 0, 0, 1);
      $button = new Gtk::Button(' OK ');
      $button->signal_connect('clicked', \&config_ok, $elements);
      $elements->{ok} = $button;
      $hbox->pack_end($button, 0, 0, 1);
      
      

      $window->show_all;
}

sub config_apply {
      my $widget = shift;
      my $elements = shift;
      my $data = $elements->{data};
      my $checkbook = $data->{checkbook};
      my $prefs = $checkbook->{prefs};

      my ($date, $sortatstart, $xsize, $ysize, $datafile);
      my ($need_reload, $update_display);

      if($elements->{DMY}->active) { $date = 'DMY'; }
      if($elements->{DYM}->active) { $date = 'DYM'; }
      if($elements->{MDY}->active) { $date = 'MDY'; }
      if($elements->{MYD}->active) { $date = 'MYD'; }
      if($elements->{YMD}->active) { $date = 'YMD'; }
      if($elements->{YDM}->active) { $date = 'YDM'; }
      if($elements->{DM}->active) { $date = 'DM'; }
      if($elements->{MD}->active) { $date = 'MD'; }

      if($date ne $prefs->{date}) {
	    $update_display = 1;
      } else {
	    $update_display = 0;
      }
      
      $prefs->{date} = $date;
      $prefs->{sortatstart} = $elements->{sortatstart}->active?'yes':'no';
      $prefs->{datafile} = $elements->{datafile}->get_text();
      $prefs->{xsize} = $elements->{xsize}->get_text();
      $prefs->{ysize} = $elements->{ysize}->get_text();

      if($update_display) {
	    my $i;
	    for($i = 0; $i < scalar(@{$checkbook->{checkbooks}}); $i++) {
		  display_checkbook($checkbook->{checkbooks}[$i],
				    $data->{clists}[$i]->{clist},
				    $prefs);
	    }
      }
      
}

sub config_save {
      my $widget = shift;
      my $elements = shift;
      my $data = $elements->{data};
      my $checkbook = $data->{checkbook};
      my $prefs = $checkbook->{prefs};

      $checkbook->save_prefs();

      my ($window, $box, $label, $button);

      $window = new Gtk::Window('toplevel');
      $window->set_position(-mouse);
      $window->set_title('Preferences Saved');
      
      $box = new Gtk::VBox(0, 0);
      $window->add($box);
      
      $label = new Gtk::Label('Preferences Saved.');
      $box->pack_start($label, 0, 0, 0);
      
      $button = new Gtk::Button('Close');
      $button->signal_connect('clicked', sub { $window->destroy(); }, $window);
      $box->pack_start($button, 0, 0, 0);


      $window->show_all;
}

sub config_ok {
      my $widget = shift;
      my $elements = shift;

      config_apply($widget, $elements);
      config_save($widget, $elements);

      $elements->{window}->destroy;
}

sub config_help {
            my $button = shift;
      my $data = shift;

      my $window = new Gtk::Window('toplevel');
      $window->set_title('About Checkbook');
      $window->position(-center);
      my $vbox = new Gtk::VBox(0, 0);
      my $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 1, 1, 1);
      my $text = new Gtk::Text(undef, undef);
      $text->set_word_wrap(1);
      $text->set_usize(Gtk::Gdk::Font::string_width($window->style->font,
"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"), -1);
#      my $label = new Gtk::Label(<<EOTXT);
      $text->freeze();
      $text->insert(undef, undef, undef, <<EOTXT);
Ok - Apply & Save
Apply - apply the settings for this session.
Save - Save the settings for future sessions.
Close - Close the window (any unapplied or unsaved changes will be lost).

Date Format - This controls the format that is used to display the date in your 
    checkbook.  The date is always internally stored in the format YYYY/MM/DD
    for efficiency reasons, but you can have it displayed in nearly any way
    that you want.

Starting Size - This is the size (in pixels) that you want this program to 
    start up in.  This is just a convenience because you can always resize
    the window to any size that you want.  This is just for making your life
    easier at startup.

Sort at Startup - You can either sort the entries in your checkbook when
    this program starts, or you can trust that they are in the correct order.
    If you do not have a habit of editing the data file by hand, you can
    uncheck this box without a problem, this is only really for people who
    make a habit of editing their checkbook by hand when this program is not
    running.

Datafile - This is simply the name of the file to use for storing your 
    checkbook.  There is no really good reason to change this, but if
    some reason I could not think of arises, feel free to take advantage
    of this option.
EOTXT
      $text->thaw();
      $hbox->pack_start($text, 1, 1, 1);
      my $scroll = new Gtk::VScrollbar($text->vadj);
      $hbox->pack_start($scroll, 0, 0, 0);
      my $button = new Gtk::Button('Close');
      $button->signal_connect('clicked', sub {$window->destroy}, $window);

      $window->add($vbox);
#      $vbox->pack_start($label, 0, 0, 0);
      $vbox->pack_start($button, 0, 0, 0);

      $window->show_all;
      1;
}


sub getting_started_box {
      my $button = shift;
      my $data = shift;

      my $window = new Gtk::Window('toplevel');
      $window->set_title('About Checkbook');
      $window->position(-center);
      my $vbox = new Gtk::VBox(0, 0);
      my $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 1, 1, 1);
      my $text = new Gtk::Text(undef, undef);
      $text->set_word_wrap(1);
      $text->set_usize(Gtk::Gdk::Font::string_width($window->style->font,
"This is a line of text that is exactly eighty characters long, with enough luck."), -1);
#      my $label = new Gtk::Label(<<EOTXT);
      $text->freeze();
      $text->insert(undef, undef, undef, <<EOTXT);
        If you\'re just getting started, then most likely you don\'t see anything below the toolbar.  Don\'t worry, just go to operations->create register and make a new register. This will start to make the program look like the screenshots.
	Next you can start adding new entries with the insert command. I suggest making the starting balance a deposit, even if it technically wasn\'t a deposit, as this will achieve what you\'re looking for.
	Make sure to save before you exit.  There is no automatic save feature in this application.  If you really want one, email me and let me know.  There probably will be one once I get preferences set up.
	  The preferences are stored in the file ~/.checkbook_preferences.  The format is documented in that file.  Your checkbook entries are normally stored in ~/.checkbook where each line is an entry and fields are separated by three colons (:::).  The date is stored internally in the form YYYY/MM/DD for faster sorting.  If you plan to make any changes to this please respect that format or you will screw things up.  If you want to import a checkbook, just run the command `cat checkbook_to_import >> ~/.checkbook` and that will graft the old checkbook onto your checkbook the next time that you open your checkbook.
EOTXT
      $text->thaw();
      $hbox->pack_start($text, 1, 1, 1);
      my $scroll = new Gtk::VScrollbar($text->vadj);
      $hbox->pack_start($scroll, 0, 0, 0);
      my $button = new Gtk::Button('Close');
      $button->signal_connect('clicked', sub {$window->destroy}, $window);

      $window->add($vbox);
#      $vbox->pack_start($label, 0, 0, 0);
      $vbox->pack_start($button, 0, 0, 0);

      $window->show_all;
      1;
}

sub start_box {
      my $button = shift;
      my $data = shift;

      my $window = new Gtk::Window('toplevel');
      $window->set_title('About Checkbook');
      $window->position(-center);
      my $vbox = new Gtk::VBox(0, 0);
      my $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 1, 1, 1);
      my $text = new Gtk::Text(undef, undef);
      $text->set_word_wrap(1);
      $text->set_usize(Gtk::Gdk::Font::string_width($window->style->font,
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"), 400);
#      my $label = new Gtk::Label(<<EOTXT);
      $text->freeze();
      $text->insert(undex, undef, undef, <<EOTXT);
The following is the license that this software is provided under.  You must
agree to it or you must return this package to whomever you purchased it from
for a full refund.  Considering that you probably got it for free, that is
not too harsh a stipulation.

While not the whole of the license by any means, the part that I am especially
concerned with is the fact that unless you have already negotiated a contract
for support with me, the author (Christopher Lansdown), there is no warranty
of any kind supplied with this program as you never provided me with any
remuneration for such a warranty.  Local laws may not allow this, but if they
do not, they are wrong.  If you plan to use this software, you may either
trust my anecdotal evidence that it has worked for me so far without any bugs
that I have noticed, or you may read the source code to this application 
yourself (since it is written in perl, simply read the executable and the 
associated module in your favorite text editor to ensure that this program is
bug free.

If this program contains bugs, I am sorry.  If you email me I will try to find
them and correct them.  However, I would be doing this completely out of the 
goodness of my heart and you must be aware that I neither guarantee that I will
try nor that if I do I will be successful.  You probably have not paid me
anything for this program so you have no right to expect anything from it.
Please do not.

Whether or not you have paid me for a support contract, please always
sanity-check all data generated by this program.  If something seems 
unreasonable, it probably is.

Anyhow, you got this for free so I guarantee nothing to you.  If you want 
guarantees, email me and we will discuss making up a support contract
that includes remuneration for my labor and risk.  Until then, observe
the wise persian proverb, "Trust in God, but tie your camel tight."

The following is the GPL, the license that I am granting you with the 
distribution of this program:


EOTXT
      $text->insert(undef, undef, undef, $data->{checkbook}->gpl);
      $text->thaw();
      $hbox->pack_start($text, 1, 1, 1);
      my $scroll = new Gtk::VScrollbar($text->vadj);
      $hbox->pack_start($scroll, 0, 0, 0);
      my $button = new Gtk::Button('Close');
      $button->signal_connect('clicked', sub {$window->destroy}, $window);

      $window->add($vbox);
#      $vbox->pack_start($label, 0, 0, 0);
      $vbox->pack_start($button, 0, 0, 0);

      $window->show_all;
      1;
}

sub license_box {
      my $button = shift;
      my $data = shift;

      my $window = new Gtk::Window('toplevel');
      $window->set_title('About Checkbook');
      $window->position(-center);
      my $vbox = new Gtk::VBox(0, 0);
      my $hbox = new Gtk::HBox(0, 0);
      $vbox->pack_start($hbox, 1, 1, 1);
      my $text = new Gtk::Text(undef, undef);
      $text->set_word_wrap(1);
      $text->set_usize(Gtk::Gdk::Font::string_width($window->style->font,
"XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"), 400);
#      my $label = new Gtk::Label(<<EOTXT);
      $text->freeze();
      $text->insert(undef, undef, undef, $data->{checkbook}->gpl);
      $text->thaw();
      $hbox->pack_start($text, 1, 1, 1);
      my $scroll = new Gtk::VScrollbar($text->vadj);
      $hbox->pack_start($scroll, 0, 0, 0);
      my $button = new Gtk::Button('Close');
      $button->signal_connect('clicked', sub {$window->destroy}, $window);

      $window->add($vbox);
#      $vbox->pack_start($label, 0, 0, 0);
      $vbox->pack_start($button, 0, 0, 0);

      $window->show_all;
      1;
}

sub about_box {
      my $button = shift;
      my $data = shift;

      my $window = new Gtk::Window('toplevel');
      $window->set_title('About Checkbook');
      $window->position(-center);
      my $vbox = new Gtk::VBox(0, 0);
      my $label = new Gtk::Label(<<EOTXT);
This program was is copyright Christopher Lansdown
and distributed under the GPL.  See the source for
more information.

I wrote this program because doing math by hand is 
just a pain.  Erasing is even worse.  I hope that
you find it useful.  Please send all bug reports and
patches to lansdoct\@cs.alfred.edu.
EOTXT
;

      my $button = new Gtk::Button('Close');
      $button->signal_connect('clicked', sub {$window->destroy}, $window);

      $window->add($vbox);
      $vbox->pack_start($label, 0, 0, 0);
      $vbox->pack_start($button, 0, 0, 0);

      $window->show_all;
      1;
}

sub null {
      return 1;
}
